// This is a modified Graph.cpp for coderange & pointer parsing

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "insx86.hpp"

#ifndef  __GRAPHNODE
#define  __GRAPHNODE

#define  TRUE  1
#define  FALSE 0

class GraphNode
{
private:
   class GraphNode   **link_down;   // pointer to an array of "children"
                                    // links
   unsigned int      num_link_down; // number of downlinks in this node                  class GraphNode   **link_up;     // pointer to an array of "parent"
                                    // links
   class GraphNode   **link_up;     // pointer to an array of "parent"
   unsigned int      num_link_up;   // number of uplinks in this node
   char              *node_name;    // pointer to the node name
   char              *node_text;    // additional text to fill into
                                    // the node
   unsigned int      ea_begin;      // begin of code block
   unsigned int      ea_end;        // end of code block
   unsigned int      vert_order;     // Vertical level for graph formatting
public:
   GraphNode();   // constructor
   ~GraphNode();  // destructor
   int   GetLinkNumDown(){return num_link_down;};  // Get Number of downlinks
   int   GetLinkNumUp(){return num_link_up;};   // Get Number of uplinks
   int   AddDownLink(class GraphNode *child);   
   int   AddUpLink(class GraphNode *parent);
   int   DelDownLink(class GraphNode *child);
   int   DelUpLink(class GraphNode *parent);   
   void  SetNodeName(char *nuname);
   void  SetNodeText(char *text);
   void  SetBegin(unsigned int begin){ea_begin = begin;}
   void  SetEnd(unsigned int end){ea_end = end;}
   void  SetVertical(unsigned int order){vert_order = order;}
         
   unsigned int GetVertical(){return vert_order;}
   unsigned int GetBegin(){return ea_begin;}
   unsigned int GetEnd(){return ea_end;}
   const char *GetNodeText(){return node_text;}
   const char *GetNodeName(){return node_name;}
   GraphNode *GetDownLink(unsigned int idx);
   GraphNode *GetUpLink(unsigned int idx);
};

GraphNode::GraphNode()
{
   link_down = 0;
   link_up = 0;
   
   num_link_down = 0;
   num_link_up = 0;   

   node_name = 0;
   node_text = 0;
   
   ea_begin = 0;
   ea_end = 0;
   vert_order = -1;
}

GraphNode *GraphNode::GetDownLink(unsigned int idx)
{
   if(num_link_down != 0)
      return(link_down[idx-1]);  
   else
      return(NULL);
}

GraphNode *GraphNode::GetUpLink(unsigned int idx)
{
   if(num_link_up != 0)
      return(link_up[idx-1]);  
   else
      return(NULL);
}

void GraphNode::SetNodeName(char *nuname)
{
   if(node_name != 0)
      qfree(node_name);
   node_name = (char *)qalloc(strlen(nuname)+1);
   strncpy(node_name, nuname, strlen(nuname)+1);
}

void GraphNode::SetNodeText(char *text)
{
   if(node_text != 0)
      qfree(node_text);
   node_text = (char *)qalloc(strlen(text)+1);
   strncpy(node_text, text, strlen(text)+1);
}

// unlinks the node from all others, qfrees memory
// Insanely slow. Really insanely slow.
GraphNode::~GraphNode() 
{
   unsigned int   i;
   
   for(i = 0; i < num_link_up; i++)
      DelUpLink(link_up[i]);
   
   for(i = 0; i < num_link_down; i++)
      DelDownLink(link_down[i]);   
}

// This function is going to be painfully slow due to the qrealloc
// upon every node addition and due to unnecessare calling upwards & downwards
// in the graph...ahwell, fuckit. I will improve performance
// at a later point.
// These functions should never ever return FALSE (!), something insanely
// wrong must've happened if they do. 

GraphNode::AddDownLink(class GraphNode *child)
{
   unsigned int   i;
   
   // first check if a Downlink to the child is already there
   for(i = 0; i < num_link_down; i++)
   {
      if(child == link_down[i])
         return(TRUE);     // link exists already, return TRUE
   }
   if(num_link_down == 0)
      link_down = (GraphNode **)qalloc(++num_link_down * sizeof(GraphNode *));   
   else
      link_down = (GraphNode **)qrealloc(link_down, ++num_link_down * sizeof(GraphNode *));
   
   if(link_down == 0)
      return(FALSE);    // out of memory ??
   link_down[num_link_down-1] = child;   
   
   if(child->AddUpLink(this) == TRUE)
      return(TRUE);  // Child successfully added uplink
   else
      return(FALSE);
}

GraphNode::AddUpLink(class GraphNode *parent)
{
   unsigned int   i;
   
   // first check if a Downlink to the child is already there
   for(i = 0; i < num_link_up; i++)
   {
      if(parent == link_up[i])
         return(TRUE);     // link exists already, return TRUE
   }
   if(num_link_up == 0)
      link_up = (GraphNode **)qalloc(++num_link_up * sizeof(GraphNode *));   
   else
      link_up = (GraphNode **)qrealloc(link_up, ++num_link_up * sizeof(GraphNode *));

   if(link_up == 0)
      return(FALSE);    // out of memory ??
   link_up[num_link_up-1] = parent;   
      
   if(parent->AddDownLink(this) == TRUE)
      return(TRUE);  // Parent successfully added downlink
   else
      return(FALSE);
}

GraphNode::DelDownLink(class GraphNode *child)
{
   unsigned int   i;
   
   for(i = 0; i < num_link_down; i++)
   {
      if(child == link_down[i])
      {
         link_down[i] = 0;
         memmove(&link_down[i], &link_down[i+1], (num_link_down-i-1) * sizeof(GraphNode *));
         num_link_down--;
         if(num_link_down != 0)
            link_down = (GraphNode **)qrealloc(link_down, num_link_down * sizeof(GraphNode *));
         else
         {            
            qfree(link_down);
            link_down = 0;
         }
         
         if(child->DelUpLink(this) == TRUE)  // delete the uplink    
            return TRUE;
         else
            return FALSE; 
      }
   }
   return(TRUE);   
}

GraphNode::DelUpLink(class GraphNode *parent)
{
   unsigned int   i;
   
   for(i = 0; i < num_link_up; i++)
   {
      if(parent == link_up[i])
      {
         link_up[i] = 0;
         memmove(&link_up[i], &link_up[i+1], (num_link_up-i-1) * sizeof(GraphNode *));
         num_link_up--;
         if(num_link_up != 0)
            link_up = (GraphNode **)qrealloc(link_up, num_link_up * sizeof(GraphNode *));
         else
         {
            qfree(link_up);
            link_up = 0;
         }
         
         if(parent->DelDownLink(this) == TRUE)  // delete the uplink    
            return TRUE;
         else
            return FALSE; 
      }
   }
   return(TRUE);   
}

class Graph
{
private:
   unsigned int   numnodes;
   GraphNode      **allNodes;    // The first node is the root node
   char           *xlate_ida2grph(char *line);
                  // Translates IDA's color scheme to the graphing plugin
public:
   Graph();
   ~Graph();
   GraphNode *AddNode(GraphNode *from, char *name);
   void OutputWingraphFile(char *filename);
   GraphNode *GetRootNode(){return allNodes[0];}
   GraphNode *FindNodeByName(char *name);
   GraphNode *FindNodeByBegin(unsigned int ea_begin);
   void Graph::SortByBegin();
};

Graph::Graph()
{
   numnodes = 1;
   allNodes = (GraphNode **)qalloc(sizeof(GraphNode *));
   allNodes[0] = new(GraphNode);
   allNodes[0]->SetNodeName("root");
}

Graph::~Graph()
{
   unsigned int i;
   
   for(i = 0; i < numnodes; i++)
      allNodes[i]->~GraphNode();   

   qfree(allNodes);      
}

void clean_idaline(char *line)
{
   unsigned int idx1, idx2;
   
   idx1 = 0;
   idx2 = 0;
   
   while(line[idx2] != 0)
   {
      if(line[idx2] == ';')      // BEGIN UGLY HACK TO MAKE SHIT WORK
      {
         while(line[idx2] != 0)  
         {
            line[idx2] = 0;
            idx2++;   
         }                       
         break;
      }                       // END UGLY HACK TO MAKE SHIT WORK
      
      if((line[idx2] == 0x01)||(line[idx2] == 0x02))
         idx2 += 2;
      else
      {
         if(line[idx2] == '\"')
         {
            line[idx1++] = '\\';
            line[idx1++] = '\"';
            idx2++;
         }
         else
            line[idx1++] = line[idx2++];   
      }
   }
   line[idx1] = 0;
   strcat(line, "\n");
}

                  // Translates IDA's color scheme to the graphing plugin

// at this point it sucks not to have any CS education. How the hell 
// do I sort something quickly ? THIS HAS TO BE REPLACED BY SOMETHING
// MORE EFFICIENT QUICKLY
void Graph::SortByBegin()
{
   unsigned int   i, tmp, highest_beg, c;

   for(i = 0; i < numnodes; i++)
      allNodes[i]->SetVertical(-1); // Set all Verticals to -1
   
   msg("sorting");   
   for(i = numnodes-1; i != 0; i--)
   {
      highest_beg = 0;
      for(c = 0; c < numnodes; c++)
      {
         if(allNodes[c]->GetVertical() == -1)
         {
            if(allNodes[c]->GetBegin() > highest_beg)
               highest_beg = allNodes[c]->GetBegin();         
         }
      }  // now we have the node with the highest order
      for(tmp = 0; tmp < numnodes; tmp++)
      {
         if(allNodes[tmp]->GetBegin() == highest_beg)
            allNodes[tmp]->SetVertical(i);   // give current vertical
      }           
   }  
   msg("...done\n");
}

GraphNode *Graph::FindNodeByBegin(unsigned int ea_begin)
{
   unsigned int i;
   
   for(i = 0; i < numnodes; i++)
   {
      if(allNodes[i]->GetBegin() == ea_begin)
         return(allNodes[i]);   
   }
   return(NULL);
}

GraphNode *Graph::FindNodeByName(char *name)
{
   unsigned int i;
   
   for(i = 0; i < numnodes; i++)
   {
      if(strcmp(name, allNodes[i]->GetNodeName()) == 0)
         return(allNodes[i]);   
   }
   return(NULL);      
}

GraphNode *Graph::AddNode(GraphNode *from, char *name)
{
   numnodes++;
   allNodes = (GraphNode **)qrealloc(allNodes, sizeof(GraphNode *) * numnodes);
   allNodes[numnodes-1] = new(GraphNode);
   allNodes[numnodes-1]->SetNodeName(name);   
   from->AddDownLink(allNodes[numnodes-1]);
   return(allNodes[numnodes-1]);
}

void Graph::OutputWingraphFile(char *filename)
{
   FILE  *fpout;
   char  asmline[MAXSTR];
   char  header[] = {
      "graph: {\n"
      "title: \"Graph Visualization\"\n"
      "manhattan_edges: yes\n"
      "layoutalgorithm: mindepth\n"
      "finetuning: no\n"
      "layout_downfactor: 100\n"
      "layout_upfactor: 0\n"
      "layout_nearfactor: 0\n"
      "xlspace: 12\n"
      "yspace: 30\n"};
   unsigned int   i, i2, numlinks;
   GraphNode      *linked_node;
   ea_t           linecount;
     
   fpout = fopen(filename, "w+");
   
   fprintf(fpout, header);
   for(i = 1; i < numnodes; i++) // skip rootnode
   {
      numlinks = allNodes[i]->GetLinkNumDown();
      fprintf(fpout, "node: { ");
      fprintf(fpout, "title: \"%s\" ", allNodes[i]->GetNodeName());
      fprintf(fpout, "label: \"\n");
      fprintf(fpout, "NodeBegin: %lx\n", allNodes[i]->GetBegin());

      linecount = allNodes[i]->GetBegin();
      while(linecount <= allNodes[i]->GetEnd())
      {
         ua_ana0(linecount);
         generate_disasm_line(linecount, asmline);
         clean_idaline(asmline);
         fprintf(fpout, "%08lx:\t\t%s", linecount, asmline);
         linecount += cmd.size;
      }

      fprintf(fpout, "NodeEnd:   %lx\n", allNodes[i]->GetEnd());
      if(allNodes[i]->GetNodeText() != 0)
         fprintf(fpout, "%s\n", allNodes[i]->GetNodeText());

      fprintf(fpout, "\"\n");   // node description end
      if(allNodes[i]->GetVertical() != -1)
         fprintf(fpout, "vertical_order: %d\n", allNodes[i]->GetVertical());
      
      fprintf(fpout, "}\n");
         
      for(i2 = 0; i2 < numlinks; i2++)
      {
         linked_node = allNodes[i]->GetDownLink(i2+1);
         fprintf(fpout, "edge: { sourcename: \"%s\" targetname: \"%s\" }\n", allNodes[i]->GetNodeName(), linked_node->GetNodeName());
      }
   }
   fprintf(fpout, "}\n");   
   fclose(fpout);      
}

#endif