/* 
   A simple plugin that creates a textfile "final.txt" that contains
   a flowgraph of the function at screenea. More a C++ coding exercise
   than anything else.
   It is an improvement over the builtin flowgraph representation as
   it includes out-of-function references as well (tremendous help in
   IIS disassembling)
   The major disadvantage is a lack of TRUE/FALSE parsing in the links
   both in color
*/
 
#include <ida.hpp>
#include <idp.hpp>
#include <loader.hpp>
#include <kernwin.hpp>
#include <frame.hpp>
#include <funcs.hpp> 
#include <xref.hpp>
#include <bytes.hpp>
#include <nalt.hpp>
#include <ua.hpp>
#include <struct.hpp>
#include <bytes.hpp>
#include <typeinf.hpp>
#include "insx86.hpp"   // home-grown documentation of cmd.itype, actually
                        // there is a "real" version along from ilfak, in
                        // the future might even use it ;>
#include "x86grph.hpp"
#include "ptrgraph.cpp"    // pointer access graph class

void  AddFlow(Graph *Graph, GraphNode *from, ea_t begin);

int __stdcall init(void)
{
   msg("...initializing %s v%d.%d plugin...", PROGRAM_NAME, MAJOR_VERSION, MINOR_VERSION);

   if(strcmp(inf.procName, "metapc") != 0)
   {
      msg("  CPU Type should be \"metapc\" but is %s ! Aborting ...\n", inf.procName);
      return PLUGIN_SKIP;
   }
   msg("  ... init successful !\n");
   return PLUGIN_OK;
}

//--------------------------------------------------------------------------
//      terminate

void __stdcall term(void)
{
}
//--------------------------------------------------------------------------
//      Creating the code block/access graph

Graph *CreateFlowGraph(ea_t begin)
{
   Graph       *Flow;
   GraphNode   *root;
   
   Flow = new(Graph);   // new graph
   root = Flow->GetRootNode();
   
   AddFlow(Flow, root, begin);
   return(Flow);
}

#define DEBUG
#ifdef DEBUG
   int   DbgFileCount = 0;
   char  DbgFileName[20];
#endif

void  AddFlow(Graph *Graph, GraphNode *from, ea_t begin)
{
   char           name[20];
   GraphNode      *curr_node;  
   ea_t           currinstr = begin, xref_to, xref_from;
   unsigned int   isize;
   
   curr_node = Graph->FindNodeByBegin(currinstr);
   if(curr_node)                       // This node exists, create a new xref
   {                                   // and break out of it
      from->AddDownLink(curr_node);    
      return;
   }
         
   sprintf(name, "%08lx", begin);
   curr_node = Graph->AddNode(from, name);
   curr_node->SetBegin(begin);

#ifdef DEBUG
//   sprintf(DbgFileName, "%08lx.txt", DbgFileCount++);
//   Graph->OutputWingraphFile(DbgFileName);
#endif

   while(currinstr != BADADDR)
   {
      ua_ana0(currinstr);           // current instruction
      isize = cmd.size;
      if(isize == 0)
         isize = 1;

      xref_from = get_first_fcref_from(currinstr);

      if(xref_from != BADADDR)   // this would be way nicer if we had lastXR :(
      {
         if((cmd.itype == x86_jmpA)||
            (cmd.itype == x86_jmpB)||
            (cmd.itype == x86_jmpC)||
            (cmd.itype == x86_jmp_short))
         {
            curr_node->SetEnd(currinstr);
            AddFlow(Graph, curr_node, get_first_fcref_from(currinstr));
            return;
         }
         else if( (cmd.itype == x86_callA)||
                  (cmd.itype == x86_callB)||
                  (cmd.itype == x86_callC))
         {
            msg("call encountered, no node added\n");
         }
         else  // everything else is a conditional branch
         {
            curr_node->SetEnd(currinstr);
            AddFlow(Graph, curr_node, get_first_fcref_from(currinstr));
            AddFlow(Graph, curr_node, get_first_cref_from(currinstr));
            return;
         }                     
      }
      if(cmd.itype == x86_retn)
      {
         curr_node->SetEnd(currinstr);
         return;    
      }
      xref_to = get_first_fcref_to(currinstr+isize);  // next instructions
                                                      // xrefs
      if(xref_to != BADADDR)
      {
         curr_node->SetEnd(currinstr);
         AddFlow(Graph, curr_node, currinstr+isize);
         return;
      }
      currinstr += isize;
   }
}

//--------------------------------------------------------------------------
//      The plugin method
void __stdcall run(int arg)
{
   Graph *grph;
   msg("Running the plugin...\n");   
   grph = CreateFlowGraph(get_screen_ea());

   grph->OutputWingraphFile("unsort.vcg");
   grph->SortByBegin();
   grph->OutputWingraphFile("final.vcg");
#ifdef DEBUG
   DbgFileCount = 0;       
#endif
}

//--------------------------------------------------------------------------
char comment[] = {PROGRAM_NAME "/metapc"};
                  
char help[] =  HELP_STRING;
                
//--------------------------------------------------------------------------
// This is the preferred name of the plugin module in the menu system
// The preferred name may be overriden in plugins.cfg file

char wanted_name[] = {PROGRAM_NAME "/metapc"};
                     
// This is the preferred hotkey for the plugin module
// The preferred hotkey may be overriden in plugins.cfg file
// Note: IDA won't tell you if the hotkey is not correct
//       It will just disable the hotkey.

char wanted_hotkey[] = WANTED_KEY;

//--------------------------------------------------------------------------
//
//      PLUGIN DESCRIPTION BLOCK
//
//--------------------------------------------------------------------------

extern "C" plugin_t PLUGIN = {
  IDP_INTERFACE_VERSION,
  0,                    // plugin flags
  init,                 // initialize
  term,                 // terminate. this pointer may be NULL.
  run,                  // invoke plugin
  comment,              // long comment about the plugin
                        // it could appear in the status line
                        // or as a hint
  help,                 // multiline help about the plugin
  wanted_name,          // the preferred short name of the plugin
  wanted_hotkey         // the preferred hotkey to run the plugin
};
